// hash.h

#ifndef _HASH_H_
#define _HASH_H_

#include <string>
#include <list>

#include "hash_h.h"
#include "CsTokenTypes.hpp"
#include "HashString.h"
//#include "allocator_extern.h"
#include "common_macros.h"
//#include "macro.h"
//#include "config.h"

using namespace HashString;

//----------------------------------------------------------


//----------------------------------------------------------


//----------------------------------------------------------

//#if MACRO_KEYWORD_UNION && !MACRO_DEPENDENCY
//#error
//#endif

//----------------------------------------------------------

void hash_init();

hash_code_type evalHashCode ( const char * s );
hash_code_type evalHashCode ( const char * s_begin, const char * s_end );
hash_code_type evalHashCode ( const char * s_begin, unsigned & length );

//----------------------------------------------------------
// mk_flag - macro/keyword flag
//----------------------------------------------------------

//#if MACRO_DEPENDENCY
//
//enum mk_flag
//{
//	mk_keyword = 0x01, // true - keyword, false - macro
//	mk_macro = 0x02, // true - keyword, false - macro
//	mk_identifier_used = 0x04, // imported macro, keyword or identifier (not defined as macro in this file) used in file
//	mk_keyword_redefined_by_macro = 0x08, // keyword was redefined by macro - just info flag
//	mk_macro_imported = 0x10, // true - macro defined in imported file, false - macro defined in this file
//	mk_macro_defined = 0x20, // macro was undefined by #undef directive
//	mk_macro_undefined = 0x40, // macro was undefined by #undef directive
////	mk_macro_absent = 0x80, // macro absent, hash record present, macro absent (identifier or undefined macro)
//	mk_keyword_and_macro = 0x100, // keyword and macro in one record
//	mk_macro_import_export = 0x200, // macro imported and redefined (exported)
//};
//
//struct KeywordMacro
//{
//	CppTokenType keyword_type;
//	PMacro macro;
//};
//
//struct TwoMacros
//{
//	PMacro imported_macro;
//	PMacro exported_macro;
//};
//
//#endif // MACRO_DEPENDENCY

//----------------------------------------------------------

//#if MACRO_KEYWORD_UNION
//
//union MacroKeywordUnion
//{
//	PMacro macro;
//	CppTokenType keyword_type;
//#if MACRO_DEPENDENCY
//	KeywordMacro * keyword_macro;
//	TwoMacros * two_macros;
//#endif
//
//	explicit MacroKeywordUnion()
//		: keyword_type ( CppTokenType(0) )
//	{}
//
//	explicit MacroKeywordUnion ( PMacro macro_ )
//		: macro(macro_)
//	{}
//
//	explicit MacroKeywordUnion ( CppTokenType keyword_type_ )
//		: keyword_type(keyword_type_)
//	{}
//};
//
//#endif // MACRO_KEYWORD_UNION

//----------------------------------------------------------
// MacroHashTable
//----------------------------------------------------------

class MacroHashTable
{
//	friend class FileCache;
public:
	struct hash_id_struct;
	typedef hash_id_struct* p_hash_id_struct;

public:
	MacroHashTable();
	~MacroHashTable();

	byte * allocate ( size_t size );
	void deallocate ( byte * first, byte * end );

	void init();
	void init ( const MacroHashTable & rhs );
	void init_empty();
	void set_zero();
	void clear();
	//inline unsigned getHashTableSize() const;
	//static inline unsigned evalHashIndex ( WPHashString );
	inline unsigned getHashTableSize() const { return hash_table_size_; }
	inline unsigned getHashTableCapacity() const { return hash_table_capacity_; }
	inline unsigned evalHashIndex ( WPHashString phs ) const
	{
		return getHashCode ( phs ) >> hash_code_shift_right_;	
	}
//	void setCapacityAs ( const MacroHashTable & rhs );
	void grow2();
	void add_macro (
		WPHashString id
		);
//	bool add_keyword ( WPHashString id, CppTokenType token_type );
//#if MACRO_KEYWORD_UNION
//	MacroKeywordUnion
//#else
//	PMacro
//#endif
//		get_macro_or_keyword ( WPHashString id, bool keyword_only, bool & keyword );
//	PMacro get_macro ( WPHashString id );
//#if MACRO_DEPENDENCY
	bool is_macro_defined (
		WPHashString id
		) const;
//#else
//	bool is_macro_defined ( WPHashString id, PMacro *& macro_ref ) const;
//#endif
//	CppTokenType get_keyword ( WPHashString id ) const;
	void delete_macro (
		WPHashString id
		);
//#if MACRO_DEPENDENCY
//	void replace_macro ( PMacro new_macro, hash_id_struct * record );
//#else
//	void replace_macro ( PMacro new_macro, PMacro * macro_ref );
//#endif



// statistics & print

	unsigned calcTotalIds() const;
	unsigned calcTotalSquareCoef() const;

	void printFull() const;
	void printStatistic() const;
//#if MACRO_DEPENDENCY
//	static void printKeywordFlags ( int flag );
//	static void printMacroFlags ( int flag );
//	void printMacroInfo() const;
//#endif

	struct hash_id_struct
	{
//#if MACRO_DEPENDENCY
//		int flag_;
//#else
//		bool keyword_;
//#endif // MACRO_DEPENDENCY
		PHashString id_; // key
		hash_id_struct * next_;
//#if MACRO_KEYWORD_UNION
//		MacroKeywordUnion union_;
//#else
//		void * data_;
//#endif

//#if MACRO_KEYWORD_UNION
//		hash_id_struct ( int flag, WPHashString id, MacroKeywordUnion union__ )
//			: flag_(flag), id_(id), next_(0), union_(union__) {}
//#else
	//#if MACRO_DEPENDENCY
		//hash_id_struct ( /*int flag,*/ WPHashString id/*, void * data*/ )
		//	: /*flag_(flag),*/ id_(id), next_(0)/*, data_(data)*/ {}
		hash_id_struct (
			WPHashString id
			)
			: id_(id)
			, next_(0)
		{}
	//#else
	//	hash_id_struct ( bool keyword, WPHashString id, void * data )
	//		: keyword_(keyword), id_(id), next_(0), data_(data) {}
	//#endif // MACRO_DEPENDENCY
//#endif

		static inline void * operator new ( size_t size );
		static inline void operator delete ( void * address );
	};

private:
	static const unsigned begin_hash_table_capacity2 = 3;
	static const unsigned begin_hash_table_capacity = ( 1 << begin_hash_table_capacity2 );
	const unsigned f_hash_code_shift_right ( unsigned capacity2 ) { return 32 - capacity2; }
public:
	hash_id_struct ** entries_;
private:
	unsigned hash_table_capacity_; // number of entries
	unsigned hash_code_shift_right_;
	unsigned hash_table_size_;
};

//----------------------------------------------------------
// HashIdStructAllocator
//----------------------------------------------------------


typedef FixedSmallObjAllocator
	< ALIGN_TO_2_POW_N_OR_64 ( sizeof ( MacroHashTable::hash_id_struct ) ), KeywordMacroHashTableStructAllocatorId, OwnPageAllocator >
		MacroHashTableStructAllocator;

//----------------------------------------------------------

inline void * MacroHashTable::hash_id_struct::operator new ( size_t size )
{
	return MacroHashTableStructAllocator::allocate();
}

//----------------------------------------------------------

inline void MacroHashTable::hash_id_struct::operator delete ( void * address )
{
	MacroHashTableStructAllocator::deallocate ( address );
}


//----------------------------------------------------------
// KeywordHashTable
//----------------------------------------------------------

class KeywordHashTable
{
public:
	struct hash_id_struct;

public:
	KeywordHashTable();
	~KeywordHashTable();

	void init();
	void init_empty();
	void set_zero();
	void clear();
	//inline unsigned getHashTableSize() const;
	//static inline unsigned evalHashIndex ( WPHashString );
	inline unsigned getHashTableSize() const { return hash_table_size_; }
	inline unsigned getHashTableCapacity() const { return hash_table_capacity_; }
	inline unsigned evalHashIndex ( WPHashString phs ) const
	{
		return getHashCode ( phs ) >> hash_code_shift_right_;	
	}
//	void setCapacityAs ( const KeywordHashTable & rhs );
//	void grow2();
	bool add_keyword ( WPHashString id, CsTokenType token_type );
	CsTokenType	get_keyword ( WPHashString id ) const;


//#ifdef _DEBUG
//	bool equal ( const KeywordHashTable & rhs ) const;
//	bool cache_equal ( const KeywordHashTable & rhs ) const;
//#endif

// statistics & print

	unsigned calcTotalIds() const;
	unsigned calcTotalSquareCoef() const;

	void printFull() const;
	void printStatistic() const;
//#if MACRO_DEPENDENCY
//	static void printKeywordFlags ( int flag );
//	static void printMacroFlags ( int flag );
//	void printMacroInfo() const;
//#endif

	struct hash_id_struct
	{
		PHashString id_; // key
		CsTokenType token_type_; // value
		hash_id_struct * next_;

		hash_id_struct ( WPHashString id, CsTokenType token_type_ )
			: id_(id), next_(0), token_type_(token_type_) {}

		static inline void * operator new ( size_t size );
		static inline void operator delete ( void * address );
	};

private:
	static const unsigned begin_hash_table_capacity2 = 8;
	static const unsigned begin_hash_table_capacity = ( 1 << begin_hash_table_capacity2 );
	const unsigned f_hash_code_shift_right ( unsigned capacity2 ) { return 32 - capacity2; }
	hash_id_struct ** entries_;
	unsigned hash_table_capacity_; // number of entries
	unsigned hash_code_shift_right_;
	unsigned hash_table_size_;
};

//----------------------------------------------------------
// KeywordIdStructAllocator
//----------------------------------------------------------


typedef FixedSmallObjAllocator
	< ALIGN_TO_2_POW_N_OR_64 ( sizeof ( KeywordHashTable::hash_id_struct ) ), KeywordMacroHashTableStructAllocatorId, OwnPageAllocator >
		KeywordIdStructAllocator;

//----------------------------------------------------------

inline void * KeywordHashTable::hash_id_struct::operator new ( size_t size )
{
	return KeywordIdStructAllocator::allocate();
}

//----------------------------------------------------------

inline void KeywordHashTable::hash_id_struct::operator delete ( void * address )
{
	KeywordIdStructAllocator::deallocate ( address );
}


//----------------------------------------------------------
// ScopeHashTable
//----------------------------------------------------------


enum entity_search_result
{
	entity_found = 0,
	entity_not_found = 1,
	entity_ambiguous = 2,
};

struct CsEntity;

class ScopeHashTable
{
	friend class CsDisplay;

public:
	struct hash_id_struct;
	struct hash_id_struct2;

public:
	ScopeHashTable();
	explicit ScopeHashTable ( unsigned power_of_2 );
	~ScopeHashTable();

	byte * allocate ( size_t size );
	void deallocate ( byte * first, byte * end );

	void set_zero();
	void clear();
	inline unsigned getHashTableSize() const { return hash_table_size_; }
	inline unsigned getHashTableCapacity() const { return hash_table_capacity_; }
	inline unsigned evalHashIndex ( WPHashString phs ) const
	{
		return getHashCode ( phs ) >> hash_code_shift_right_;	
	}
	void grow2();
	bool add_id ( WPHashString id, CsEntity * data );
	//void add_id ( WPHashString id, CsEntity * data );
	//bool add_unique_id ( WPHashString id, CsEntity * data );
	//bool add_shared_id ( WPHashString id, CsEntity * data );
	entity_search_result get_id ( WPHashString id, CsEntity *& entity ) const;
	CsEntity * get_single_id ( WPHashString id ) const;
	//std::pair<const hash_id_struct2*,const hash_id_struct2*> get_id_list ( WPHashString id ) const;
	const hash_id_struct2 * get_id_list ( WPHashString id ) const;
	//bool get_id ( WPHashString id, CsEntity *& entity ) const;
	//bool get_ids ( WPHashString id, std::list<CsEntity*> & list ) const;
	//CsEntity * get_type ( WPHashString id ) const;
	bool exist_id ( WPHashString id ) const;
	//void delete_id ( WPHashString id );
	void delete_entities();
	//void reserve ( unsigned size );


	// statistics & print


	struct hash_id_struct2
	{
		CsEntity * data_;
		hash_id_struct2 * next_;

		hash_id_struct2()
			: data_(0), next_(0) {}
		hash_id_struct2 ( CsEntity * data )
			: data_(data), next_(0) {}
		hash_id_struct2 ( CsEntity * data, hash_id_struct2 * next )
			: data_(data), next_(next) {}

		static inline void * operator new ( size_t size );
		static inline void operator delete ( void * address );
	};

	struct hash_id_struct
	{
		PHashString id_; // key
		hash_id_struct * next_;
		//CsEntity * data_;
		hash_id_struct2 data_;

		hash_id_struct()
			: id_(0), next_(0), data_(0) {}
			hash_id_struct ( WPHashString id, CsEntity * data )
			: id_(id), next_(0), data_ ( data ) {}

		static inline void * operator new ( size_t size );
		static inline void operator delete ( void * address );
	};

	typedef hash_id_struct* p_hash_id_struct;
	typedef hash_id_struct2* p_hash_id_struct2;

private:
	static const unsigned begin_hash_table_capacity2 = 4;
	static const unsigned begin_hash_table_capacity = ( 1 << begin_hash_table_capacity2 );
	const unsigned f_hash_code_shift_right ( unsigned capacity2 ) { return 32 - capacity2; }
public:
	hash_id_struct ** entries_;
	unsigned hash_table_capacity_; // number of entries
	unsigned hash_code_shift_right_;
	unsigned hash_table_size_;
};

//----------------------------------------------------------
// ScopeHashIdStructAllocator
//----------------------------------------------------------


typedef FixedSmallObjAllocator
	< ALIGN_TO_2_POW_N_OR_64 ( sizeof ( ScopeHashTable::hash_id_struct ) ), ScopeHashTableStructAllocatorId, OwnPageAllocator >
		ScopeHashIdStructAllocator;

//----------------------------------------------------------

inline void * ScopeHashTable::hash_id_struct::operator new ( size_t size )
{
	return ScopeHashIdStructAllocator::allocate();
}

//----------------------------------------------------------

inline void ScopeHashTable::hash_id_struct::operator delete ( void * address )
{
	ScopeHashIdStructAllocator::deallocate ( address );
}


//----------------------------------------------------------
// ScopeHashIdStructAllocator2
//----------------------------------------------------------


typedef FixedSmallObjAllocator
	< ALIGN_TO_2_POW_N_OR_64 ( sizeof ( ScopeHashTable::hash_id_struct2 ) ), ScopeHashTableStructAllocatorId, OwnPageAllocator >
		ScopeHashIdStructAllocator2;

//----------------------------------------------------------

inline void * ScopeHashTable::hash_id_struct2::operator new ( size_t size )
{
	return ScopeHashIdStructAllocator2::allocate();
}

//----------------------------------------------------------

inline void ScopeHashTable::hash_id_struct2::operator delete ( void * address )
{
	ScopeHashIdStructAllocator2::deallocate ( address );
}



//----------------------------------------------------------
// IntScopeHashTable
//----------------------------------------------------------


struct CsScopeEntity;

class IntScopeHashTable
{
public:
	IntScopeHashTable();
	~IntScopeHashTable();

	byte * allocate ( size_t size );
	void deallocate ( byte * first, byte * end );

	void set_zero();
	void clear();
	inline unsigned getHashTableSize() const { return hash_table_size_; }
	inline unsigned getHashTableCapacity() const { return hash_table_capacity_; }
	inline unsigned evalHashIndex ( unsigned id ) const;
	void grow2();
	//void add_id ( WPHashString id, CsEntity * data );
	bool add_unique_id ( unsigned id, CsScopeEntity * data );
	//bool add_shared_id ( WPHashString id, CsEntity * data );
	bool get_id ( unsigned id, CsScopeEntity *& entity ) const;
	//bool get_ids ( WPHashString id, std::list<CsEntity*> & list ) const;
	//CsEntity * get_type ( WPHashString id ) const;
	//bool exist_id ( WPHashString id ) const;
	//void delete_id ( WPHashString id );


	// statistics & print

	unsigned calcTotalIds() const;
	unsigned calcTotalSquareCoef() const;

	void printFull() const;
	void printStatistic() const;

	struct hash_id_struct
	{
		unsigned id_; // key
		hash_id_struct * next_;
		CsScopeEntity * data_;

		hash_id_struct()
			: id_(0), next_(0), data_(0) {}
		hash_id_struct ( unsigned id, CsScopeEntity * data )
			: id_(id), next_(0), data_(data) {}

		static inline void * operator new ( size_t size );
		static inline void operator delete ( void * address );
	};
	typedef hash_id_struct* p_hash_id_struct;

private:
	static const unsigned begin_hash_table_capacity2 = 4;
	static const unsigned begin_hash_table_capacity = ( 1 << begin_hash_table_capacity2 );
	const unsigned f_hash_code_shift_right ( unsigned capacity2 ) { return 32 - capacity2; }
	hash_id_struct ** entries_;
	unsigned hash_table_capacity_; // number of entries
	unsigned hash_code_shift_right_;
	unsigned hash_table_size_;
};

//----------------------------------------------------------
// IntScopeHashIdStructAllocator
//----------------------------------------------------------


typedef FixedSmallObjAllocator
	< ALIGN_TO_2_POW_N_OR_64 ( sizeof ( IntScopeHashTable::hash_id_struct ) ), IntScopeHashIdStructAllocatorId, OwnPageAllocator >
		IntScopeHashIdStructAllocator;

//----------------------------------------------------------

inline void * IntScopeHashTable::hash_id_struct::operator new ( size_t size )
{
	return IntScopeHashIdStructAllocator::allocate();
}

//----------------------------------------------------------

inline void IntScopeHashTable::hash_id_struct::operator delete ( void * address )
{
	IntScopeHashIdStructAllocator::deallocate ( address );
}



//----------------------------------------------------------

#endif // __HASH_H__
